@@ -10,6 +10,7 @@ import io
 import itertools
 import operator
 import os
+import random
 import re
 import sys
 import textwrap
@@ -73,7 +74,7 @@ class SiPM:
 ##############################################################################
 
 
-def identify_sipm_status(dbi, classification, wafer_pid, sequential, no_b_grade):
+def identify_sipm_status(dbi, classification, wafer_pid, sequential, no_b_grade, qcid=None, nounusablesipmcheck=False):
     """
     For the given wafer, obtain two sets of wafer (column, row) pairs, for
     good/bad SiPMs respectively. Chooses between two methods of accomplishing
@@ -85,7 +86,7 @@ def identify_sipm_status(dbi, classification, wafer_pid, sequential, no_b_grade)
     indicate whether SiPMs were deemed acceptable for use in production. This
     should be true again at some point in the future.
 
-    Currently a more complex lookup is required. For a given SiPM ID, only the
+    Currently, a more complex lookup is required. For a given SiPM ID, only the
     row with the largest value of sipm_test.sipm_qc_id is considered.
     From that row, for the SiPM to be regarded as good it must have
     sipm_test.classification == 'good' and sipm_test.quality_flag == 0.
@@ -100,19 +101,35 @@ def identify_sipm_status(dbi, classification, wafer_pid, sequential, no_b_grade)
         wafer_pid : int
         sequential : bool
         no_b_grade : bool
+        qcid : bool
+        nounusablesipmcheck : bool
     --------------------------------------------------------------------------
-    returns (set, set)
+    returns (set, set, set) or None
     --------------------------------------------------------------------------
     """
-    dfr = dbi.get('sipm', wafer_id=wafer_pid).data
+    response = dbi.get('sipm', wafer_id=wafer_pid)
+    if response.network_timeout:
+        print('WARNING: network timeout')
+        return None
+
+    dfr = response.data
+
+    if dfr.empty:
+        print('WARNING: empty response')
+        return None
 
     if classification:
         return sipm_status_by_classification(dbi, dfr)
 
-    if sequential:
-        return sipm_status_full_check(dbi, dfr, no_b_grade)
+    if nounusablesipmcheck:
+        unusable_sipm_ids = set()
+    else:
+        unusable_sipm_ids = set(dbi.get('unusable_sipm').data.sipm_id.values)
 
-    return sipm_status_full_check_async(dbi, dfr, no_b_grade)
+    if sequential:
+        return sipm_status_full_check(dbi, dfr, unusable_sipm_ids, no_b_grade, qcid)
+
+    return sipm_status_whole_table(dbi, dfr, unusable_sipm_ids, no_b_grade, qcid)
 
 
 def sipm_status_by_classification(dbi, dfr):
@@ -122,7 +139,7 @@ def sipm_status_by_classification(dbi, dfr):
 
     Uses just sipm_test.classification for the evaluation.
 
-    Note that this check as currently implemented is overly pessamistic. If a
+    Note that this check as currently implemented is overly pessimistic. If a
     SiPM ID has ever been assigned a bad classification, it will show as bad
     here. In Q4 2023, <1% of SiPM IDs had multiple entries, and a later
     corrected entry may contain a good result.
@@ -153,13 +170,13 @@ def sipm_status_by_classification(dbi, dfr):
     return wafer_map_good, wafer_map_b_grade, wafer_map_bad
 
 
-def identify_b_grade_sipms(dbi, dfr, good_sipm_ids):
+def identify_b_grade_sipms(dbi, dfr, good_sipm_ids, qcid, sipm_test):
     """
     SK's additional checks based on device performance.
 
     Here we are returning generally good SiPMs. These passed electrical and
     visual inspection at LFoundry, and passed electrical tests at NOA.
-    However their performance is at the margins of acceptable performance.
+    However, their performance is at the margins of acceptable performance.
 
     These SiPMs will not be discarded, they will be retained and handled as
     production silicon, except that they will not be used in production
@@ -172,13 +189,18 @@ def identify_b_grade_sipms(dbi, dfr, good_sipm_ids):
             with the database.
         dfr : Pandas DataFrame
         good_sipm_ids : set of int
+        qcid : int
+        sipm_test : pandas DataFrame
     --------------------------------------------------------------------------
     returns set of tuple
-        e.g. {(int, int), ...}
+        e.g. {(int, int), ...} or None
         notionally these are B-grade SiPMs
     --------------------------------------------------------------------------
     """
-    sipm = SiPMCheck(dbi)
+    try:
+        sipm = SiPMCheck(dbi, qcid, sipm_test)
+    except (LookupError, TimeoutError):
+        return None
 
     b_grade_sipm_ids = {
         sipm_id
@@ -193,7 +215,7 @@ def identify_b_grade_sipms(dbi, dfr, good_sipm_ids):
     }
 
 
-def wafer_map_sets(dbi, dfr, good_sipm_ids, no_b_grade):
+def wafer_map_sets(dbi, dfr, good_sipm_ids, no_b_grade, qcid, sipm_test=None):
     """
     Generate the sets of locations necessary to generate the wafer map.
 
@@ -205,6 +227,8 @@ def wafer_map_sets(dbi, dfr, good_sipm_ids, no_b_grade):
         dfr : Pandas DataFrame
         good_sipm_ids : set of int
         no_b_grade : bool
+        qcid : int
+        sipm_test : pandas DataFrame
     --------------------------------------------------------------------------
     returns (set of tuple, set of tuple, set of tuple)
         where each tuple is a pair of column, row values
@@ -220,7 +244,10 @@ def wafer_map_sets(dbi, dfr, good_sipm_ids, no_b_grade):
     if no_b_grade:
         wafer_map_b_grade = set()
     else:
-        wafer_map_b_grade = identify_b_grade_sipms(dbi, dfr, good_sipm_ids)
+        wafer_map_b_grade = identify_b_grade_sipms(dbi, dfr, good_sipm_ids, qcid, sipm_test)
+        if wafer_map_b_grade is None:
+            wafer_map_b_grade = set()
+            print('ERROR: could not read status of b-grade SiPMs')
 
     wafer_map_good -= wafer_map_b_grade
 
@@ -228,15 +255,12 @@ def wafer_map_sets(dbi, dfr, good_sipm_ids, no_b_grade):
     wafer_map_bad = all_locations - wafer_map_good - wafer_map_b_grade
 
     if not wafer_map_good:
-        print(
-            'WARNING: no good SiPMs found. '
-            'If this is an older wafer, consider using --classification'
-        )
+        print('WARNING: no good SiPMs found.')
 
     return wafer_map_good, wafer_map_b_grade, wafer_map_bad
 
 
-def sipm_status_full_check(dbi, dfr, no_b_grade):
+def sipm_status_full_check(dbi, dfr, unusable_sipm_ids, no_b_grade, qcid):
     """
     For the given wafer, obtain two sets of wafer (column, row) pairs, for
     good/bad SiPMs respectively.
@@ -250,23 +274,32 @@ def sipm_status_full_check(dbi, dfr, no_b_grade):
             Instance of the Database interface class; allows communication
             with the database.
         dfr : Pandas DataFrame
+        unusable_sipm_ids : set of int
         no_b_grade : bool
+        qcid : int
     --------------------------------------------------------------------------
-    returns (set, set)
+    returns (set of tuple, set of tuple, set of tuple)
+        where each tuple is a pair of column, row values
     --------------------------------------------------------------------------
     """
     # this will give us 268 SiPM IDs, not 264
     all_sipms_ids_for_wafer = set(dfr.sipm_pid.values)
 
-    columns = ['classification', 'quality_flag', 'sipm_qc_id']
+    columns = ['classification', 'quality_flag', 'sipm_qc_id', 'timestamp']
 
     good_sipm_ids = set()
     for sipm_id in all_sipms_ids_for_wafer:
-        dfr_tmp = dbi.get('sipm_test', sipm_id=sipm_id).data
+        dfr_all_qc = dbi.get('sipm_test', sipm_id=sipm_id).data
+        if qcid is not None:
+            dfr_tmp = dfr_all_qc.loc[dfr_all_qc['sipm_qc_id'] == qcid]
+        else:
+            dfr_tmp = dfr_all_qc
 
         # Get columns for row with highest sipm_qc_id value.
         try:
-            classification, quality_flag, _ = dfr_tmp[columns].sort_values('sipm_qc_id').values[-1]
+            classification, quality_flag, *_ = dfr_tmp[columns].sort_values(
+                ['sipm_qc_id', 'timestamp'], ascending=[True, True]
+            ).values[-1]
         except IndexError:
             # We will see IndexError for the four SiPMs at the far left/right
             # edges that are not tested.
@@ -275,7 +308,71 @@ def sipm_status_full_check(dbi, dfr, no_b_grade):
             if classification == 'good' and quality_flag == 0:
                 good_sipm_ids.add(sipm_id)
 
-    return wafer_map_sets(dbi, dfr, good_sipm_ids, no_b_grade)
+    good_sipm_ids -= unusable_sipm_ids
+
+    return wafer_map_sets(dbi, dfr, good_sipm_ids, no_b_grade, qcid)
+
+
+def sipm_status_whole_table(dbi, dfr, unusable_sipm_ids, no_b_grade, qcid):
+    """
+    For the given wafer, obtain two sets of wafer (column, row) pairs, for
+    good/bad SiPMs respectively.
+
+    Uses sipm_test.classification, sipm_test.quality_flag and
+    sipm_test.sipm_qc_id for the evaluation.
+
+    Class SiPMCheck needs the entire sipm_test table. We need a subset of that
+    table here, so it's more efficient just to load the whole table here and
+    pass it to SiPMCheck.
+
+    --------------------------------------------------------------------------
+    args
+        dbi : ds20kdb.interface.Database
+            Instance of the Database interface class; allows communication
+            with the database.
+        dfr : Pandas DataFrame
+        unusable_sipm_ids : set of int
+        no_b_grade : bool
+        qcid : int
+    --------------------------------------------------------------------------
+    returns (set of tuple, set of tuple, set of tuple) or None
+        where each tuple is a pair of column, row values
+    --------------------------------------------------------------------------
+    """
+    # this will give us 268 SiPM IDs, not 264
+    all_sipms_ids_for_wafer = set(dfr.sipm_pid.values)
+
+    response = dbi.get('sipm_test')
+    if response.network_timeout:
+        print('Network timeout')
+        return None
+    if response.data is None:
+        print('No response')
+        return None
+    sipm_test = response.data
+
+    columns = ['classification', 'quality_flag', 'sipm_qc_id', 'timestamp']
+
+    good_sipm_ids = set()
+    for sipm_id in all_sipms_ids_for_wafer:
+        dfr_tmp = sipm_test[sipm_test['sipm_id'] == sipm_id]
+
+        # Get columns for row with highest sipm_qc_id value.
+        try:
+            classification, quality_flag, *_ = dfr_tmp[columns].sort_values(
+                ['sipm_qc_id', 'timestamp'], ascending=[True, True]
+            ).values[-1]
+        except IndexError:
+            # We will see IndexError for the four SiPMs at the far left/right
+            # edges that are not tested.
+            pass
+        else:
+            if classification == 'good' and quality_flag == 0:
+                good_sipm_ids.add(sipm_id)
+
+    good_sipm_ids -= unusable_sipm_ids
+
+    return wafer_map_sets(dbi, dfr, good_sipm_ids, no_b_grade, qcid, sipm_test)
 
 
 async def fetch(url, session, sipm_id):
@@ -303,7 +400,7 @@ async def fetch(url, session, sipm_id):
                 response.raise_for_status()
                 data = await response.text()
         except aiohttp.ClientError:
-            await asyncio.sleep(1)
+            await asyncio.sleep(1 + random.random() * 4)
 
     return {sipm_id: data}
 
@@ -313,7 +410,7 @@ async def fetch_all(dbi, urls):
     Fetch database responses for all SiPM URLs.
 
     The database cannot cope with the default upper limit of 100 concurrent
-    connections, so limit this to something it can handle. Note that the the
+    connections, so limit this to something it can handle. Note that the
     limit set here needs to take into consideration others connecting to the
     database. Too many concurrent connections causes the database to generate
     500 Internal Server Error.
@@ -365,7 +462,7 @@ def csv_to_dataframe(response):
     )
 
 
-def sipm_status_full_check_async(dbi, dfr, no_b_grade):
+def sipm_status_full_check_async(dbi, dfr, unusable_sipm_ids, no_b_grade, qcid):
     """
     For the given wafer, obtain two sets of wafer (column, row) pairs, for
     good/bad SiPMs respectively.
@@ -379,9 +476,12 @@ def sipm_status_full_check_async(dbi, dfr, no_b_grade):
             Instance of the Database interface class; allows communication
             with the database.
         dfr : Pandas DataFrame
+        unusable_sipm_ids : set of int
         no_b_grade : bool
+        qcid : int
     --------------------------------------------------------------------------
-    returns (set, set)
+    returns (set of tuple, set of tuple, set of tuple) or None
+        where each tuple is a pair of column, row values
     --------------------------------------------------------------------------
     """
     # this will give us 268 SiPM IDs, not 264
@@ -397,23 +497,34 @@ def sipm_status_full_check_async(dbi, dfr, no_b_grade):
     # SiPM) to form a single dict.
     mapping = functools.reduce(operator.ior, asyncio.run(fetch_all(dbi, urls)), {})
 
-    columns = ['classification', 'quality_flag', 'sipm_qc_id']
+    columns = ['classification', 'quality_flag', 'sipm_qc_id', 'timestamp']
     good_sipm_ids = set()
     for sipm_id, csv_str in mapping.items():
-        dfr_tmp = csv_to_dataframe(csv_str)
+        dfr_all_qc = csv_to_dataframe(csv_str)
+        if qcid is not None:
+            dfr_tmp = dfr_all_qc.loc[dfr_all_qc['sipm_qc_id'] == qcid]
+        else:
+            dfr_tmp = dfr_all_qc
 
         # Get columns for row with highest sipm_qc_id value.
         try:
-            classification, quality_flag, _ = dfr_tmp[columns].sort_values('sipm_qc_id').values[-1]
+            classification, quality_flag, *_ = dfr_tmp[columns].sort_values(
+                ['sipm_qc_id', 'timestamp'], ascending=[True, True]
+            ).values[-1]
         except IndexError:
             # We will see IndexError for the four SiPMs at the far left/right
             # edges that are not tested.
             pass
+        except KeyError:
+            print(f'ERROR: broken response for sipm_id {sipm_id}')
+            return None
         else:
             if classification == 'good' and quality_flag == 0:
                 good_sipm_ids.add(sipm_id)
 
-    return wafer_map_sets(dbi, dfr, good_sipm_ids, no_b_grade)
+    good_sipm_ids -= unusable_sipm_ids
+
+    return wafer_map_sets(dbi, dfr, good_sipm_ids, no_b_grade, qcid)
 
 
 ##############################################################################
@@ -426,17 +537,30 @@ class SiPMCheck:
     Class of functions to check a sipm test values against the NOA test
     distributions.
     """
-    def __init__(self, dbi):
+    def __init__(self, dbi, qcid=None, sipm_test=None):
         """
-        loading the database separetely to allow the following functions to be
+        loading the database separately to allow the following functions to be
         able to loop over multiple sipms. If more tables are needed can
         thread the loading and save in a list
         """
         self.dbi = dbi
-        self.df_sipm_qc = self.dbi.get('sipm_test').data
 
-        # Get Latest QA/QC criteria
-        self.current_sipm_qc = self.df_sipm_qc['sipm_qc_id'].max()
+        # no need to load the table if we already have it in memory
+        if sipm_test is None:
+            response = self.dbi.get('sipm_test')
+            if response.network_timeout:
+                raise TimeoutError
+            if response.data is None:
+                raise LookupError
+            self.df_sipm_qc = response.data
+        else:
+            self.df_sipm_qc = sipm_test
+
+        if qcid is None:
+            # Get Latest QA/QC criteria
+            self.current_sipm_qc = self.df_sipm_qc['sipm_qc_id'].max()
+        else:
+            self.current_sipm_qc = qcid
 
         # Get NOA test results for sipms passing the latest quality qc cuts
         self.df_sipm_good = self.df_sipm_qc.loc[
@@ -449,7 +573,7 @@ class SiPMCheck:
         """
         Simple function to make a CDF of the NOA Distributions. They generally
         don't seem to follow any standard analytical function, so this is a
-        bit of a work around.
+        bit of a workaround.
 
         This function looks for the x value intersect where the
         sum == 1 - pct. (i.e P(X<x) == 1-pct).
@@ -561,21 +685,26 @@ class SiPMCheck:
         return all(map(stat_test, tests))
 
 
-##############################################################################
-# identify empty sipm entries
-#
-# so when populating
-##############################################################################
-
-
 ##############################################################################
 # python 3.6 compatibility
 ##############################################################################
 
 
 def grouper(iterable, n, fillvalue=None):
-    """Collect data into fixed-length chunks or blocks"""
-    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx"
+    """
+    Collect data into fixed-length chunks or blocks
+
+    grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx"
+
+    Ref:
+    https://docs.python.org/3.6/library/itertools.html#itertools-recipes
+
+    Python 3.11 has a recipe for batched():
+    https://docs.python.org/3.11/library/itertools.html#itertools-recipes
+
+    Python 3.12 has batched() in the standard library which is similar:
+    https://docs.python.org/3.12/library/itertools.html#itertools.batched
+    """
     args = [iter(iterable)] * n
     return itertools.zip_longest(*args, fillvalue=fillvalue)
 
@@ -887,7 +1016,7 @@ def fill_file(filename, missing_sipms, sipm_allocation, lot, wafer_number, write
 
     Existing SiPM entries may be present, or missing altogether.
 
-    If the file contains any broken lines they will be propogated into the
+    If the file contains any broken lines they will be propagated into the
     new file.
 
     Try to make minimal changes to the file, to avoid creating confusing
@@ -994,7 +1123,7 @@ def fill_files(var, lot, wafer_number, write=True):
         external file amended
     --------------------------------------------------------------------------
     """
-    sipm_type, existing_tray_files, new_sipms = var
+    _sipm_type, existing_tray_files, new_sipms = var
 
     lines_for_all_files = []
     for filename, (_contents, missing_sipms) in existing_tray_files.items():
@@ -1063,9 +1192,46 @@ def write_tray_file(lot, wafer_number, detailed, filename, write):
     return header + lines
 
 
-def create_tray_files(var, lot, wafer_number, write, directory):
+def tray_filename(sipm_type, lot, wafer_number, offset, tray_number):
     """
-    FIXME - check directory structure: 9323959/01
+    Create a properly formatted filename for green/yellow SiPM tray files,
+    e.g.
+
+    green:  9323959_04_green_tray_04.txt
+    yellow: yellow_0002.txt
+
+    Green files are created in a directory specifically related to their
+    wafer. So it makes sense to name them accordingly.
+
+    Yellow files detail the contents of SiPM trays containing devices that
+    were regarded as production standard, but failed visual inspection. These
+    SiPMs may be used later for production, but the trays they are housed in
+    may contain a devices from many different wafers, hence the generic
+    filename.
+
+    --------------------------------------------------------------------------
+    args
+        sipm_type : string
+        lot : int
+        wafer_number : int
+        offset : int
+        tray_number : int
+    --------------------------------------------------------------------------
+    returns : list
+    --------------------------------------------------------------------------
+    """
+    if sipm_type == 'green':
+        return (
+            f'{lot}_{wafer_number:02}_'
+            f'{sipm_type.lower()}_tray_{tray_number:02}.txt'
+        )
+
+    return f'yellow_{offset + tray_number:04}.txt'
+
+
+def create_tray_files(var, lot, wafer_number, write):
+    """
+    Create tray files for green or yellow SiPMs.
 
     --------------------------------------------------------------------------
     args
@@ -1077,22 +1243,18 @@ def create_tray_files(var, lot, wafer_number, write, directory):
         lot : int
         wafer_number : int
         write : book
-        directory : string
     --------------------------------------------------------------------------
     returns : none
         external file amended
     --------------------------------------------------------------------------
     """
-    sipm_type, available_sipms = var
+    sipm_type, available_sipms, directory, offset = var
 
     lines_for_all_files = []
 
     for i, x in enumerate(grouper(available_sipms, 24), start=1):
         detailed = [(t[0], t[1], lot, wafer_number) for t in x if t]
-        filename = (
-            f'{lot}_{wafer_number:02}_'
-            f'{sipm_type.lower()}_tray_{i:02}.txt'
-        )
+        filename = tray_filename(sipm_type, lot, wafer_number, offset, i)
 
         lines_for_all_files += write_tray_file(
             lot,
