@@ -12,7 +12,6 @@ Output: history of the object and post on the DB object>_location table
 Author: P.Franchini - p.franchini@lancaster.ac.uk
 """
 
-import contextlib
 import datetime
 import itertools
 import platform
@@ -248,15 +247,17 @@ class Object:
         if self.system == 'Darwin':
             # macOS Ventura 13.6.5
             return event.keycode == 17 and event.state == 16
-        elif self.system == 'Linux':
+
+        if self.system == 'Linux':
             # Pop!_OS 22.04 LTS x86_64 + Debian GNU/Linux 12 (bookworm) aarch64
             return event.keycode == 55 and event.state == 20
-        elif self.system == 'Windows':
+
+        if self.system == 'Windows':
             # MWS Windows 10
             return event.keycode == 17 and event.state == 16
-        else:
-            # ASSUME anything else is related to Linux
-            return event.keycode == 55 and event.state == 20
+
+        # ASSUME anything else is related to Linux
+        return event.keycode == 55 and event.state == 20
 
     @staticmethod
     def is_valid_date_format(input_string):
@@ -351,7 +352,7 @@ class Object:
                 try:
                     tmp_id = self.dbi.get(
                         test_object, **{key: parameter}
-                    ).data[select][0]
+                        ).data[select].iloc[-1] # gives the very last entry in case of multiple versions (eg. vPDUs)
                 except (IndexError, KeyError):
                     break
                 else:
@@ -368,7 +369,12 @@ class Object:
                 break
 
         self.error_label.config(text=status)
-        self.output_label.config(text=obj)
+        # prints the grade of a vTile
+        if obj=='vtile':
+            quality = self.dbi.get_vtile_qc(qrcode)
+            self.output_label.config(text=obj+': '+quality)
+        else:
+            self.output_label.config(text=obj)
 
         return obj, parameter
 
